'use strict';
const _ = require('lodash');
const async = require('async');
const path = require('path');
const updateJSON = require('./update-json');
const fs = require('fs');
const semver = require('./semver');
const github = require('./github');
const replaceInFile = require('replace-in-file');
const markdown = require('./markdown');
const yaml = require('yaml');
const execSync = require('child_process').execSync;
const { Octokit } = require('@octokit/rest');
const childProcess = require('child_process');
const octokit = new Octokit({
    debug: Boolean(process.env.DEBUG),
    auth: process.env.GITHUB_TOKEN,
});
const templateDefaults = [
    '{{#*inline "commits"}}',
    '{{> render-header }}',
    '{{> block-newline}}',
    '{{#each commits}}',
    '{{> render-with-nesting nesting=../nesting block=../block ~}}',
    '{{/each}}',
    '{{> block-newline}}',
    '{{/inline~}}',
    '{{#*inline "render-with-nesting"}}',
    '{{~#if this.nested ~}}',
    '{{> block-newline}}',
    '{{> block-prefix}}<details>',
    '{{> block-prefix}}<summary> {{> render-with-author-inline}} </summary>',
    '{{> block-newline}}',
    '{{#each this.nested}}',
    '{{> commits nesting=(append ../nesting "#") block=(append ../block ">") }}',
    '{{/each}}',
    '{{> block-newline}}',
    '{{> block-prefix}}</details>',
    '{{> block-newline}}',
    '{{~else~}}',
    '{{> block-prefix}}* {{> render-with-author}}',
    '{{~/if~}}',
    '{{/inline}}',
    '{{#*inline "render-with-author"}}',
    '{{#if this.author}}',
    '{{this.subject}} [{{this.author}}]',
    '{{else}}',
    '{{this.subject}}',
    '{{/if}}',
    '{{/inline}}',
    '{{#*inline "render-with-author-inline"}}',
    '{{#if this.author ~}}',
    '{{this.subject}} [{{this.author}}] ',
    '{{~else~}}',
    '{{this.subject}} ',
    '{{~/if~}}',
    '{{/inline}}',
    '{{#*inline "block-prefix"}}',
    '{{#isnt block "" ~}}{{block}} {{/isnt~}}',
    '{{/inline}}',
    '{{#*inline "block-newline"}}',
    '{{#isnt block ""}}{{block}} {{else}}{{/isnt}}',
    '{{/inline}}',
    ''
].join('\n');
const isIncrementalCommit = (changeType) => {
    return Boolean(changeType) && changeType.trim().toLowerCase() !== 'none';
};
const getAuthor = (commitHash) => {
    if (commitHash) {
        return execSync(`git show --quiet --format="%an" ${commitHash}`, {
            encoding: 'utf8',
        }).replace('\n', '');
    }
    return 'Unknown author';
};
const getChangeType = (footer) => {
    return footer[Object.keys(footer).find((k) => k.toLowerCase() === 'change-type')];
};
const extractContentsBetween = (changelog, repo, start, end) => {
    return _(changelog)
        .filter((entry) => {
        return semver.lt(start, entry.version) && semver.lte(entry.version, end);
    })
        .map((entry) => {
        entry.version = `${repo}-${entry.version}`;
        return entry;
    })
        .value();
};
const getNestedChangeLog = async (options, commit, startVersion, endVersion, callback) => {
    const { owner, repo, ref = await github.getDefaultBranch(octokit, owner, repo), } = options;
    try {
        const response = await github.getChangelogYML(owner, repo, ref, octokit);
        const changelog = yaml.parse(Buffer.from(response, 'base64').toString());
        const nested = extractContentsBetween(changelog, repo, startVersion, endVersion);
        return callback(null, nested);
    }
    catch (err) {
        return callback(new Error(`Could not find .versionbot/CHANGELOG.yml in ${repo} under branch ${ref}`));
    }
};
const resolveTag = (shash, upstream) => {
    const result = childProcess.spawnSync('git', ['describe', '--abbrev=0', shash], {
        cwd: `${process.cwd()}/${upstream}`,
    });
    return result.stdout.toString().trim();
};
const processSubmoduleOutput = (data, upstream) => {
    var lines = `${data}`.split('\n');
    for (let i = 0; i < lines.length; i++) {
        const line = lines[i];
        const regexp = new RegExp(`Submodule ${upstream} (\\S+)\\.\\.(\\S+):`);
        const match = line.match(regexp);
        if (match) {
            const previousVersion = resolveTag(match[1], upstream);
            const newVersion = resolveTag(match[2], upstream);
            return [previousVersion, newVersion];
        }
    }
};
const attachNestedChangelog = (upstreams, commit, callback) => {
    async.map(upstreams, (upstream, cb) => {
        const regexp = new RegExp(`[Uu]pdate ${upstream.pattern} from (\\S+) to (\\S+)`);
        const match = commit.body.match(regexp);
        if (match) {
            const currentVersion = match[1];
            const targetVersion = match[2];
            return getNestedChangeLog(upstream, commit, currentVersion, targetVersion, cb);
        }
        const regexp1 = new RegExp(`[Uu]pdate ${upstream.pattern}`);
        const match1 = commit.body.match(regexp1);
        if (match1) {
            const result = childProcess.spawnSync('git', [
                'log',
                '--oneline',
                '--submodule',
                '-U0',
                'HEAD^1..HEAD',
            ]);
            const response = processSubmoduleOutput(result.stdout, `${upstream.pattern}`);
            if (response) {
                const currentVersion = response[0];
                const targetVersion = response[1];
                return getNestedChangeLog(upstream, commit, currentVersion, targetVersion, cb);
            }
        }
        return cb(null);
    }, (err, nestedCommits) => {
        commit.nested = _(nestedCommits).compact().flatten().value();
        return callback(err, commit);
    });
};
const INITIAL_CHANGELOG = `# Changelog

All notable changes to this project will be documented in this file
automatically by Versionist. DO NOT EDIT THIS FILE MANUALLY!
This project adheres to [Semantic Versioning](http://semver.org/).

`;
const touchChangelog = (changelogPath, done) => {
    if (fs.existsSync(changelogPath)) {
        return done(null, changelogPath);
    }
    return fs.writeFile(changelogPath, INITIAL_CHANGELOG, (err) => {
        return done(err, changelogPath);
    });
};
const replace = (file, pattern, replacement, callback) => {
    async.waterfall([
        (done) => {
            fs.readFile(file, 'utf8', done);
        },
        (contents, done) => {
            if (pattern.test(contents)) {
                done(null, contents);
            }
            else {
                done(new Error(`Pattern does not match ${file}`));
            }
        },
        (contents, done) => {
            const updated = contents.replace(pattern, replacement);
            done(null, updated);
        },
        _.partial(fs.writeFile, file),
    ], callback);
};
const getCleanFunction = (options) => {
    _.defaults(options, {
        clean: true,
    });
    if (_.isRegExp(options.clean)) {
        return (version) => {
            return version.replace(options.clean, '');
        };
    }
    return options.clean ? semver.valid : _.identity;
};
function readJSON(filePath, callback) {
    fs.readFile(filePath, 'utf8', (error, data) => {
        let pjObj;
        if (error) {
            if (error.code === 'ENOENT') {
                error.message = `No such file or directory: ${filePath}`;
            }
        }
        else {
            try {
                pjObj = JSON.parse(data);
            }
            catch (err) {
                error = err;
            }
        }
        callback(error, pjObj);
    });
}
module.exports = {
    subjectParser: {
        angular: (options, subject) => {
            const subjectParts = subject.match(/^(?:fixup!\s*)?(\w*)(\(([\w$.*/-]*)\))?: (.*)$/);
            return {
                type: _.nth(subjectParts, 1),
                scope: _.nth(subjectParts, 3),
                title: _.nth(subjectParts, 4) || subject,
            };
        },
    },
    includeCommitWhen: {
        angular: (options, commit) => {
            if (_.isString(commit.subject)) {
                return _.some([
                    _.startsWith(commit.subject, 'feat'),
                    _.startsWith(commit.subject, 'fix'),
                    _.startsWith(commit.subject, 'perf'),
                ]);
            }
            return _.includes(['feat', 'fix', 'perf'], commit.subject.type);
        },
        'has-changetype': (options, commit) => {
            const changeType = getChangeType(commit.footer);
            return (isIncrementalCommit(changeType) ||
                isIncrementalCommit(module.exports.getIncrementLevelFromCommit.subject({}, commit)));
        },
        'has-changelog-entry': (options, commit) => {
            return Boolean(commit.footer['changelog-entry']);
        },
    },
    getIncrementLevelFromCommit: {
        'change-type': (options, commit) => {
            const changeType = getChangeType(commit.footer);
            if (isIncrementalCommit(changeType)) {
                return changeType.trim().toLowerCase();
            }
        },
        subject: (options, commit) => {
            if (!_.isString(commit.subject)) {
                return null;
            }
            const match = commit.subject.match(/^(patch|minor|major):/i);
            if (_.isArray(match) && isIncrementalCommit(match[1])) {
                return match[1].trim().toLowerCase();
            }
        },
        'change-type-or-subject': (options, commit) => {
            const ctIncrement = module.exports.getIncrementLevelFromCommit['change-type'](options, commit);
            if (_.isString(ctIncrement)) {
                return ctIncrement;
            }
            return module.exports.getIncrementLevelFromCommit.subject(options, commit);
        },
    },
    getChangelogDocumentedVersions: {
        'changelog-headers': (options, file, callback) => {
            const cleanFn = getCleanFunction(options);
            fs.readFile(file, {
                encoding: 'utf8',
            }, (error, changelog) => {
                if (error) {
                    if (error.code === 'ENOENT') {
                        return callback(null, []);
                    }
                    return callback(error);
                }
                const versions = _.chain(markdown.extractTitles(changelog))
                    .map((title) => {
                    return _.filter(_.split(title, ' '), semver.valid);
                })
                    .flattenDeep()
                    .map(cleanFn)
                    .value();
                return callback(null, versions);
            });
        },
    },
    getCurrentBaseVersion: {
        'latest-documented': (options, documentedVersions, history, callback) => {
            return callback(null, semver.getGreaterVersion(documentedVersions));
        },
    },
    addEntryToChangelog: {
        prepend: (options, file, entry, callback) => {
            _.defaults(options, {
                fromLine: 6,
            });
            async.waterfall([
                _.partial(touchChangelog, file),
                (touchedFiles, done) => {
                    fs.readFile(file, {
                        encoding: 'utf8',
                    }, done);
                },
                (contents, done) => {
                    const changelogLines = _.split(contents, '\n');
                    return done(null, _.join(_.reduce([
                        _.slice(changelogLines, 0, options.fromLine),
                        _.split(entry, '\n'),
                        _.slice(changelogLines, options.fromLine),
                    ], (accumulator, array) => {
                        const head = _.dropRightWhile(accumulator, _.isEmpty);
                        const body = _.dropWhile(array, _.isEmpty);
                        if (_.isEmpty(head)) {
                            return body;
                        }
                        return _.concat(head, [''], body);
                    }, []), '\n'));
                },
                _.partial(fs.writeFile, file),
            ], callback);
        },
    },
    addEntryToHistoryFile: {
        'yml-prepend': (options, file, raw, callback) => {
            if (fs.existsSync(file)) {
                let changelog = [];
                try {
                    changelog = yaml.parse(fs.readFileSync(file, 'utf8'));
                    if (!changelog) {
                        changelog = [];
                    }
                }
                catch (e) {
                    if (e.code !== 'ENOENT') {
                        return callback(e);
                    }
                }
                changelog.unshift(raw);
                fs.writeFile(file, yaml.stringify(changelog), callback);
            }
            else {
                return callback(null);
            }
        },
    },
    transformTemplateDataAsync: {
        'nested-changelogs': (options, data, callback) => {
            async.map(data.commits, _.partial(attachNestedChangelog, options.upstream), (err, commits) => {
                if (err) {
                    return callback(err);
                }
                data.commits = _.sortBy(commits, (commit) => {
                    return Boolean(commit.nested);
                });
                return callback(err, data);
            });
        },
    },
    transformTemplateData: {
        'changelog-entry': (options, data) => {
            if (_.isEmpty(data.commits)) {
                throw new Error('All commits were filtered out for this version');
            }
            data.commits.forEach((commit) => {
                if (commit.footer) {
                    commit.subject = commit.footer['changelog-entry'] || commit.subject;
                }
                commit.author = getAuthor(commit.hash);
            });
            return data;
        },
    },
    getGitReferenceFromVersion: {
        'v-prefix': (options, version) => {
            if (_.startsWith(version, 'v')) {
                return version;
            }
            return `v${version}`;
        },
    },
    updateVersion: {
        npm: (options, cwd, version, callback) => {
            const cleanFn = getCleanFunction(options);
            const packageJSON = path.join(cwd, 'package.json');
            const packageLockJSON = path.join(cwd, 'package-lock.json');
            const npmShrinkwrapJSON = path.join(cwd, 'npm-shrinkwrap.json');
            const cleanedVersion = cleanFn(version);
            if (!cleanedVersion) {
                return callback(new Error(`Invalid version: ${version}`));
            }
            const updateNestedVersion = (manifestJSON, done) => readJSON(manifestJSON, (readError, pljObj) => {
                if (readError && readError.code === 'ENOENT') {
                    return done(null);
                }
                if (pljObj.lockfileVersion >= 2 &&
                    pljObj.packages &&
                    pljObj.packages[''] &&
                    pljObj.packages[''].version) {
                    pljObj.packages[''].version = cleanedVersion;
                    updateJSON(manifestJSON, {
                        packages: pljObj.packages,
                    }, (updateError) => {
                        if (updateError && updateError.code === 'ENOENT') {
                            return done(null);
                        }
                        return done(updateError);
                    });
                }
                else {
                    return done(null);
                }
            });
            async.waterfall([
                (done) => readJSON(packageJSON, done),
                (pjObj, done) => {
                    const publishedAt = new Date().toISOString();
                    updateJSON(packageJSON, {
                        version: cleanedVersion,
                        versionist: Object.assign(Object.assign({}, pjObj.versionist), { publishedAt }),
                    }, done);
                },
                (done) => {
                    updateJSON(packageLockJSON, {
                        version: cleanedVersion,
                    }, (error) => {
                        if (error && error.code === 'ENOENT') {
                            return done(null);
                        }
                        return done(error);
                    });
                },
                (done) => updateNestedVersion(packageLockJSON, done),
                (done) => {
                    updateJSON(npmShrinkwrapJSON, {
                        version: cleanedVersion,
                    }, (error) => {
                        if (error && error.code === 'ENOENT') {
                            return done(null);
                        }
                        return done(error);
                    });
                },
                (done) => updateNestedVersion(npmShrinkwrapJSON, done),
            ], callback);
        },
        cargo: (options, cwd, version, callback) => {
            const cleanFn = getCleanFunction(options);
            const cargoToml = path.join(cwd, 'Cargo.toml');
            const cargoLock = path.join(cwd, 'Cargo.lock');
            const cleanedVersion = cleanFn(version);
            if (!cleanedVersion) {
                return callback(new Error(`Invalid version: ${version}`));
            }
            async.waterfall([
                (done) => {
                    return fs.readFile(cargoToml, 'utf8', done);
                },
                (contents, done) => {
                    const matches = contents.match(/\[package\][^[]+?name\s*=\s*("|')(.+?)\1/m);
                    if (_.isNull(matches)) {
                        done(new Error(`Package name not found in ${cargoToml}`));
                    }
                    else {
                        done(null, matches[2]);
                    }
                },
                (packageName, done) => {
                    if (fs.existsSync(cargoLock)) {
                        replace(cargoLock, new RegExp(`(name\\s*=\\s*(?:"|')${packageName}(?:"|')[^[]+?version\\s*=\\s*)("|').*?\\2`, 'm'), '$1$2' + cleanedVersion + '$2', (err) => {
                            return done(err || null);
                        });
                    }
                    else {
                        done(null);
                    }
                },
                (done) => {
                    replace(cargoToml, /(\[package\][^[]+?version\s*=\s*)("|').*?\2/m, '$1$2' + cleanedVersion + '$2', done);
                },
            ], callback);
        },
        initPy: (options, cwd, version, callback) => {
            _.defaults(options, {
                targetFile: '__init__.py',
            });
            const cleanFn = getCleanFunction(options);
            const initFile = path.join(cwd, options.targetFile);
            const cleanedVersion = cleanFn(version);
            if (!cleanedVersion) {
                return callback(new Error(`Invalid version: ${version}`));
            }
            replaceInFile({
                files: initFile,
                from: /(__version__\s*=\s*)('|")(?:0|[1-9]\d*)\.(?:0|[1-9]\d*)\.(?:0|[1-9]\d*)\2/g,
                to: '$1$2' + cleanedVersion + '$2',
            }, (error) => {
                if (error) {
                    return callback(error);
                }
            });
        },
        quoted: (options, cwd, version, callback) => {
            _.defaults(options, {
                baseDir: '.',
                regexFlags: '',
            });
            if (path.isAbsolute(options.baseDir)) {
                return callback(new Error("baseDir option can't be an absolute path"));
            }
            if (_.isUndefined(options.file)) {
                return callback(new Error('Missing file option'));
            }
            if (path.isAbsolute(options.file)) {
                return callback(new Error("file option can't be an absolute path"));
            }
            if (_.isUndefined(options.regex)) {
                return callback(new Error('Missing regex option'));
            }
            const updateFile = path.join(cwd, options.baseDir, options.file);
            const cleanFn = getCleanFunction(options);
            const cleanedVersion = cleanFn(version);
            if (!cleanedVersion) {
                return callback(new Error(`Invalid version: ${version}`));
            }
            const innerRegex = RegExp(options.regex);
            const combinedRegexSource = '(' + innerRegex.source + ')("|\').*?\\2';
            const combinedRegexFlags = _.join(_.uniqBy(innerRegex.flags + options.regexFlags), '');
            replace(updateFile, new RegExp(combinedRegexSource, combinedRegexFlags), '$1$2' + cleanedVersion + '$2', callback);
        },
        'update-version-file': (options, cwd, version, callback) => {
            const versionFile = path.join(cwd, 'VERSION');
            return fs.writeFile(versionFile, version, {
                flag: 'w+',
            }, callback);
        },
        mixed: (options, cwd, version, callback) => {
            const wrapped = _.map([
                module.exports.updateVersion.npm,
                module.exports.updateVersion.cargo,
                module.exports.updateVersion['update-version-file'],
                module.exports.updateVersion.initPy,
            ], (updateFn) => {
                return (cb) => {
                    return updateFn(options, cwd, version, () => {
                        return cb(null);
                    });
                };
            });
            return async.parallel(wrapped, callback);
        },
    },
    updateContract: {
        version: (options, cwd, version, callback) => {
            const cleanFn = getCleanFunction(options);
            const cleanedVersion = cleanFn(version);
            if (!cleanedVersion) {
                return callback(new Error(`Invalid version: ${version}`));
            }
            const contract = path.join(cwd, 'balena.yml');
            if (!fs.existsSync(contract)) {
                return callback(null, version);
            }
            const content = yaml.parseDocument(fs.readFileSync(contract, 'utf8'));
            content.set('version', cleanedVersion);
            fs.writeFile(contract, content.toString(), callback);
        },
    },
    incrementVersion: {
        semver: (options, version, incrementLevel) => {
            if (!semver.valid(version)) {
                throw new Error(`Invalid version: ${version}`);
            }
            const incrementedVersion = semver.inc(version, incrementLevel);
            if (!incrementedVersion) {
                throw new Error(`Invalid increment level: ${incrementLevel}`);
            }
            return incrementedVersion;
        },
    },
    template: {
        oneline: templateDefaults.concat([
            '{{#*inline "render-header"}}',
            '{{> block-prefix}}{{nesting}} {{version}} - {{moment date "Y-MM-DD"}}',
            '{{/inline}}',
            '{{> commits nesting="##" block=""}}'
        ].join('\n')),
        default: templateDefaults.concat([
            '{{~#*inline "render-header"}}',
            '{{> block-prefix}}{{nesting}} {{#eq nesting "#"}}v{{else}}{{/eq}}{{version}}',
            '{{> block-prefix}}{{nesting}}# ({{moment date "Y-MM-DD"}})',
            '{{/inline}}',
            '{{> commits nesting="#" block=""}}'
        ].join('\n'))
    },
    INITIAL_CHANGELOG: INITIAL_CHANGELOG,
};
//# sourceMappingURL=presets.js.map